.daclickable.nav-link.danavlink:hover {
	border: 1px solid;
}

/* https://codepen.io/dragontheory/pen/qvJYzz */
@import url('https://fonts.googleapis.com/css?family=Raleway:400,500,600,700,800,900');

/* article, aside, details, figcaption, figure, footer, header, hgroup, main, nav, section, summary {
    display: block;
}*/

/*.al-progress-box {
  display: block;  
  box-sizing: border-box;  
  font-family: "Raleway",sans-serif;
  text-align:center;
  display: -webkit-box;
  display: -ms-flexbox;
  display: flex;
  -ms-flex-flow: row wrap;
      flex-flow: row wrap;
}*/

.al-session-form-title {
  font-weight: bolder;
}

.radialProgressBar {
  border-radius: 50%;
  width: 2.5rem;
  height: 2.5rem;
  display: -webkit-box;
  display: -ms-flexbox;
  display: flex;
  --radial-background: #ddd;
  background: var(--radial-background);
  margin: .2rem;
}

[data-bs-theme=dark] .radialProgressBar {
  --radial-background: #212529;
}

.radialProgressBar .overlay {
  border-radius: 50%;
  width: 2.3rem;
  height: 2.3rem;
  margin: auto;
  background: #fff;
  text-align: center;
  padding-top: 10%;
}

[data-bs-theme=dark] .radialProgressBar .overlay {
  background: #212529;
}

.progress-0 {
  background-image: -webkit-linear-gradient(left, var(--radial-background) 50%, transparent 50%), -webkit-linear-gradient(left, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(90deg, var(--radial-background) 50%, transparent 50%), linear-gradient(90deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-10 {
  background-image: -webkit-linear-gradient(36deg, var(--radial-background) 50%, transparent 50%), -webkit-linear-gradient(left, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(54deg, var(--radial-background) 50%, transparent 50%), linear-gradient(90deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-20 {
  background-image: -webkit-linear-gradient(72deg, var(--radial-background) 50%, transparent 50%), -webkit-linear-gradient(left, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(18deg, var(--radial-background) 50%, transparent 50%), linear-gradient(90deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-30 {
  background-image: -webkit-linear-gradient(108deg, var(--radial-background) 50%, transparent 50%), -webkit-linear-gradient(left, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(-18deg, var(--radial-background) 50%, transparent 50%), linear-gradient(90deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-40 {
  background-image: -webkit-linear-gradient(144deg, var(--radial-background) 50%, transparent 50%), -webkit-linear-gradient(left, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(-54deg, var(--radial-background) 50%, transparent 50%), linear-gradient(90deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-50 {
  background-image: -webkit-linear-gradient(right, var(--radial-background) 50%, transparent 50%), -webkit-linear-gradient(left, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(-90deg, var(--radial-background) 50%, transparent 50%), linear-gradient(90deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-60 {
  background-image: -webkit-linear-gradient(left, #028cd5 50%, transparent 50%), -webkit-linear-gradient(36deg, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(90deg, #028cd5 50%, transparent 50%), linear-gradient(54deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-70 {
  background-image: -webkit-linear-gradient(left, #028cd5 50%, transparent 50%), -webkit-linear-gradient(72deg, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(90deg, #028cd5 50%, transparent 50%), linear-gradient(18deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-80 {
  background-image: -webkit-linear-gradient(left, #028cd5 50%, transparent 50%), -webkit-linear-gradient(108deg, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(90deg, #028cd5 50%, transparent 50%), linear-gradient(-18deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-90 {
  background-image: -webkit-linear-gradient(left, #028cd5 50%, transparent 50%), -webkit-linear-gradient(144deg, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(90deg, #028cd5 50%, transparent 50%), linear-gradient(-54deg, #028cd5 50%, var(--radial-background) 50%);
}

.progress-100 {
  background-image: -webkit-linear-gradient(left, #028cd5 50%, transparent 50%), -webkit-linear-gradient(right, #028cd5 50%, var(--radial-background) 50%);
  background-image: linear-gradient(90deg, #028cd5 50%, transparent 50%), linear-gradient(-90deg, #028cd5 50%, var(--radial-background) 50%);
}
