/*---------------------------------------------------------*/
/* Logo and title 
/*---------------------------------------------------------*/

.title-container {
  display: grid; 
  grid-template-columns: min-content min-content; 
  grid-template-rows: min-content; 
  gap: 0px 0.5em; 
  grid-template-areas: 
    "logo title"; 
  justify-items: start; 
  align-items: center; 
}
.al-logo { 
  grid-area: logo; 
}
.al-logo img {
  max-height: 40px;
}
.al-title {
  display: grid; 
  grid-template-columns: min-content; 
  grid-template-rows: 1fr 1fr; 
  gap: 0px 0px; 
  grid-template-areas: 
    "title-row-1"
    "title-row-2"; 
  justify-content: start; 
  align-content: start; 
  justify-items: start; 
  align-items: start; 
  grid-area: title; 
  line-height: 1em;
  font-family: 'Inter Bold', 'Arial';
}

.title-row-1 { 
    grid-area: title-row-1; 
    font-weight: bold;
    font-size: .7em; 
    text-transform: none; /*prevent any bootstrap theme to transform it to uppercase */
  }

.title-row-2 { 
    grid-area: title-row-2;
    font-weight: bold;
    font-size: 1em; 
    text-transform: none; /*prevent any bootstrap theme to transform it to uppercase */    
  }

@media only screen and (max-width: 800px) {
  .al-title, .al-logo {
    font-weight: normal;
    font-size: 60%;    
  }
  .al-logo {    
    border-width: 4px; 
  }  
}

.danavbar-title { max-width: unset }

/*---------------------------------------------------------*/
/* Navigation bar and footer
/*---------------------------------------------------------*/
/*.navbar {
    height: 60px;
}
*/
.da-pad-for-navbar {
    padding-top: 70px;
}
.da-top-for-navbar {
    top: 70px;
}
footer {
  text-align: right
}
footer a {
  padding-left: 2em;
}
footer a:first-child {
  padding-left: 0em;
}
/*---------------------------------------------------------*/
/* Content
/*---------------------------------------------------------*/
blockquote {
    background: #f9f9f9;
    border-left: 10px solid #ccc;
    margin: 1.5em 10px;
    padding: 0.5em 10px;
    quotes: "\201C""\201D""\2018""\2019";
    font-size: .75em;
}
blockquote:before {
    color: #ccc;
    content: open-quote;
    font-size: 4em;
    line-height: 0.1em;
    margin-right: 0.25em;
    vertical-align: -0.4em;
}
blockquote p {
    display: inline;
}
/* [FILE] image shadow */
.daicon {    
  box-shadow: 5px 5px 15px #69868e;
  border-style: solid;
  border-width: 1px;
  border-color: gray;
}

td img.daicon {
  width: 100%;
}

/* Reduce integer and number input box size */
input[type=number] {
  width: 10rem;
}
/* Widen the Submit button (button labels vary) */
.btn[type='submit'] {
  min-width: 8em;
}

/* Make the footer links and da terms slightly darker in light mode 
  and lighter in dark mode for better contrast */
footer a {
  color: #0264F7;
}

[data-bs-theme=dark] footer a {
  color: #3988FF
}

.daterm {
  color: #3b842c;
  text-decoration: dashed;
  text-decoration-line: underline;
}

[data-bs-theme=dark] .daterm {
  color: #4fa23e;
}

.daquestionbackbutton {
  --bs-btn-focus-shadow-rgb: 49,132,253;
}

/* An upstream fix for dark mode, with radio buttons and checkboxes. See https://github.com/jhpyle/docassemble/pull/718 */
[data-bs-theme=dark] .btn-light.dalabelauty {
  --bs-btn-color: #fff;
  --bs-btn-bg: #212529;
  --bs-btn-border-color: #212529;
  --bs-btn-hover-color: #fff;
  --bs-btn-hover-bg: #424649;
  --bs-btn-hover-border-color: #373b3e;
  --bs-btn-focus-shadow-rgb: 66, 70, 73;
  --bs-btn-active-color: #fff;
  --bs-btn-active-bg: #4d5154;
  --bs-btn-active-border-color: #373b3e;
  --bs-btn-active-shadow: inset 0 3px 5px rgba(0, 0, 0, 0.125);
  --bs-btn-disabled-color: #fff;
  --bs-btn-disabled-bg: #212529;
  --bs-btn-disabled-border-color: #212529;
}
